@@ -13,10 +13,7 @@
 # limitations under the License.
 
 """
-A package for solving and working with sliding tile puzzles.
-
-All submodule code is imported into the parent namespace, so there is no need
-to explicitly refer to submodules.
+A package for solving sliding tile puzzles.
 
 Examples:
     >>> from slidingpuzzle import *
@@ -68,7 +65,7 @@ from slidingpuzzle.algorithms import (
     IDA_STAR,
     IDDFS,
     compare,
-    eval_heuristic,
+    evaluate,
     get_next_states,
     search,
     swap_tiles,
