@@ -16,10 +16,14 @@
 Search algorithms
 """
 
+from typing import Optional
+
 import collections
 import heapq
 
 from slidingpuzzle.board import (
+    Board,
+    FrozenBoard,
     get_empty_yx,
     get_next_moves,
     is_solvable,
@@ -42,11 +46,11 @@ GREEDY = "greedy"
 IDA_STAR = "ida*"
 IDDFS = "iddfs"
 ALGORITHMS = (
-    A_STAR,  # f(n) = g(n) + W * h(n)
+    A_STAR,
     BEAM,
     BFS,
     DFS,
-    GREEDY,  # f(n) = h(n)
+    GREEDY,
     IDA_STAR,
     IDDFS,
 )
@@ -71,7 +75,7 @@ def get_next_states(state: State) -> list[State]:
     return next_states
 
 
-def a_star(board: tuple[list[int], ...], **kwargs) -> SearchResult:
+def a_star(board: Board, **kwargs) -> SearchResult:
     r"""
     A* heuristic search algorithm.
     Supports weights, depth bounds, and f-bounds.
@@ -92,7 +96,7 @@ def a_star(board: tuple[list[int], ...], **kwargs) -> SearchResult:
         f_bound (float): A limit on state cost. Default is :math:`\infty`.
         detect_dupes (bool): Duplicate detection (i.e. track visited states).
             Default is ``True``.
-        heuristic: A function that maps boards to an estimated distance from goal.
+        heuristic: A function that maps boards to an estimated cost-to-go.
             Default is :func:`slidingpuzzle.heuristics.manhattan_distance`.
         weight (float): A constant multiplier on heuristic evaluation
 
@@ -111,7 +115,7 @@ def a_star(board: tuple[list[int], ...], **kwargs) -> SearchResult:
     empty_pos = get_empty_yx(board)
     initial_state = State(board, empty_pos)
     unvisited = [initial_state]
-    visited = set()
+    visited: set[FrozenBoard] = set()
 
     # stats
     generated, expanded = 0, 0
@@ -145,7 +149,7 @@ def a_star(board: tuple[list[int], ...], **kwargs) -> SearchResult:
     return SearchResult(board, generated, expanded, list(unvisited), visited, None)
 
 
-def beam(board: tuple[list[int], ...], **kwargs) -> SearchResult:
+def beam(board: Board, **kwargs) -> SearchResult:
     r"""
     Beam search is a variant of breadth-first search that sorts its children using a
     heuristic function and then drops child states to match the beam width. This search
@@ -158,7 +162,7 @@ def beam(board: tuple[list[int], ...], **kwargs) -> SearchResult:
         f_bound (float): A limit on state cost. Default is :math:`\infty`.
         detect_dupes (bool): Duplicate detection (i.e. track visited states).
             Default is ``True``.
-        heuristic: A function that maps boards to an estimated distance from goal.
+        heuristic: A function that maps boards to an estimated cost-to-go.
             Default is :func:`slidingpuzzle.heuristics.manhattan_distance`.
         width (int): The beam width. Default is ``3``.
 
@@ -177,7 +181,7 @@ def beam(board: tuple[list[int], ...], **kwargs) -> SearchResult:
     empty_pos = get_empty_yx(board)
     initial_state = State(board, empty_pos)
     unvisited = collections.deque([initial_state])
-    visited = set()
+    visited: set[FrozenBoard] = set()
 
     # stats
     generated, expanded = 0, 0
@@ -213,7 +217,7 @@ def beam(board: tuple[list[int], ...], **kwargs) -> SearchResult:
     return SearchResult(board, generated, expanded, list(unvisited), visited, None)
 
 
-def bfs(board: tuple[list[int], ...], **kwargs) -> SearchResult:
+def bfs(board: Board, **kwargs) -> SearchResult:
     r"""
     Breadth-first search
 
@@ -235,7 +239,7 @@ def bfs(board: tuple[list[int], ...], **kwargs) -> SearchResult:
     empty_pos = get_empty_yx(board)
     initial_state = State(board, empty_pos)
     unvisited = collections.deque([initial_state])
-    visited = set()
+    visited: set[FrozenBoard] = set()
 
     # stats
     generated, expanded = 0, 0
@@ -267,7 +271,7 @@ def bfs(board: tuple[list[int], ...], **kwargs) -> SearchResult:
     return SearchResult(board, generated, expanded, list(unvisited), visited, None)
 
 
-def dfs(board: tuple[list[int], ...], **kwargs) -> SearchResult:
+def dfs(board: Board, **kwargs) -> SearchResult:
     r"""
     Depth-first search
 
@@ -289,7 +293,7 @@ def dfs(board: tuple[list[int], ...], **kwargs) -> SearchResult:
     empty_pos = get_empty_yx(board)
     initial_state = State(board, empty_pos)
     unvisited = [initial_state]
-    visited = set()
+    visited: set[FrozenBoard] = set()
 
     # stats
     generated, expanded = 0, 0
@@ -321,7 +325,7 @@ def dfs(board: tuple[list[int], ...], **kwargs) -> SearchResult:
     return SearchResult(board, generated, expanded, list(unvisited), visited, None)
 
 
-def greedy(board: tuple[list[int], ...], **kwargs) -> SearchResult:
+def greedy(board: Board, **kwargs) -> SearchResult:
     r"""
     Greedy best-first search. This search orders all known states using the provided
     heuristic and greedily chooses the state closest to the goal.
@@ -332,7 +336,7 @@ def greedy(board: tuple[list[int], ...], **kwargs) -> SearchResult:
         f_bound (float): A limit on state cost. Default is :math:`\infty`.
         detect_dupes (bool): Duplicate detection (i.e. track visited states).
             Default is ``True``.
-        heuristic: A function that maps boards to an estimated distance from goal.
+        heuristic: A function that maps boards to an estimated cost-to-go.
             Default is :func:`slidingpuzzle.heuristics.manhattan_distance`.
 
     Returns:
@@ -349,7 +353,7 @@ def greedy(board: tuple[list[int], ...], **kwargs) -> SearchResult:
     empty_pos = get_empty_yx(board)
     initial_state = State(board, empty_pos)
     unvisited = [initial_state]
-    visited = set()
+    visited: set[FrozenBoard] = set()
 
     # stats
     generated, expanded = 0, 0
@@ -383,7 +387,7 @@ def greedy(board: tuple[list[int], ...], **kwargs) -> SearchResult:
     return SearchResult(board, generated, expanded, list(unvisited), visited, None)
 
 
-def ida_star(board: tuple[list[int], ...], **kwargs) -> SearchResult:
+def ida_star(board: Board, **kwargs) -> SearchResult:
     r"""
     Iterative deepening A*. A depth-first search that uses an f-bound instead of depth
     to limit search. The next bound is set to the minimum increase in f-bound observed
@@ -394,7 +398,7 @@ def ida_star(board: tuple[list[int], ...], **kwargs) -> SearchResult:
         depth_bound (int): A limit to search depth. Default is :math:`\infty`.
         detect_dupes (bool): Duplicate detection (i.e. track visited states).
             Default is ``True``.
-        heuristic: A function that maps boards to an estimated distance from goal.
+        heuristic: A function that maps boards to an estimated cost-to-go.
             Default is :func:`slidingpuzzle.heuristics.manhattan_distance`.
         weight (float): A constant multiplier on heuristic evaluation.
             Default is ``1``.
@@ -411,11 +415,11 @@ def ida_star(board: tuple[list[int], ...], **kwargs) -> SearchResult:
     # initial state
     goal = new_board(len(board), len(board[0]))
     empty_pos = get_empty_yx(board)
-    bound = manhattan_distance(board)
+    bound = float(manhattan_distance(board))
     initial_state = State(board, empty_pos)
     next_bound = float("inf")
     unvisited = [initial_state]
-    visited = set()
+    visited: set[FrozenBoard] = set()
 
     # stats
     generated, expanded = 0, 0
@@ -461,7 +465,7 @@ def ida_star(board: tuple[list[int], ...], **kwargs) -> SearchResult:
     return SearchResult(board, generated, expanded, list(unvisited), visited, None)
 
 
-def iddfs(board: tuple[list[int], ...], **kwargs) -> SearchResult:
+def iddfs(board: Board, **kwargs) -> SearchResult:
     r"""
     Iterative deepening depth first search. Same as :func:`dfs`, except that the depth
     bound is incrementally increased until a solution is found.
@@ -484,7 +488,7 @@ def iddfs(board: tuple[list[int], ...], **kwargs) -> SearchResult:
     initial_state = State(board, empty_pos)
     next_bound = bound
     unvisited = [initial_state]
-    visited = set()
+    visited: set[FrozenBoard] = set()
 
     # stats
     generated, expanded = 0, 0
@@ -535,11 +539,7 @@ ALGORITHMS_MAP = {
 }
 
 
-def search(
-    board: tuple[list[int], ...],
-    alg: str = A_STAR,
-    **kwargs,
-) -> SearchResult:
+def search(board: Board, alg: str = A_STAR, **kwargs) -> SearchResult:
     r"""
     Searches for a set of moves that take the provided board state to the
     solved state.
@@ -578,33 +578,41 @@ def search(
 
     if not is_solvable(board):
         raise ValueError("The provided board is not solvable.")
-    alg = alg.strip().lower()
-    if alg not in ALGORITHMS:
+    alg_ = alg.strip().lower()
+    if alg_ not in ALGORITHMS:
         raise ValueError(f'Unknown algorithm: "{alg}"')
-    return ALGORITHMS_MAP[alg](board, **kwargs)
+    return ALGORITHMS_MAP[alg_](board, **kwargs)
 
 
-def eval_heuristic(
+def evaluate(
     h: int,
     w: int,
-    heuristic,
+    heuristic=manhattan_distance,
     algorithm=A_STAR,
     num_iters: int = 64,
     **kwargs,
 ) -> float:
     """
     Runs search on ``num_iters`` random boards. Returns the average number of nodes
-    expanded.
+    generated.
+
+    Args:
+        h: Height of the board
+        w: Width of the board
+        heuristic: Heuristic function to evaluate
+        algorithm: Search algorithm to evaluate
+        num_iters: Number of iterations to average
+        kwargs: Additional args for ``algorithm``
 
     Returns:
-        The average number of states expanded.
+        The average number of states generated.
     """
     total = 0
     for _ in range(num_iters):
         board = new_board(h, w)
         shuffle_board(board)
         result = search(board, alg=algorithm, heuristic=heuristic, **kwargs)
-        total += result.expanded
+        total += result.generated
     return total / num_iters
 
 
@@ -615,23 +623,41 @@ def compare(
     hb=manhattan_distance,
     alga=A_STAR,
     algb=A_STAR,
-    num_iters: int = 8,
+    num_iters: int = 32,
+    kwargsa: Optional[dict] = None,
+    kwargsb: Optional[dict] = None,
     **kwargs,
 ) -> tuple[float, float]:
     """
     Runs search on ``num_iters`` random boards, trying both alga(board, ha) and
-    algb(board, hb) on each board. Returns the average number of states expanded
+    algb(board, hb) on each board. Returns the average number of states generated
     for both.
 
+    Args:
+        h: Height of the board
+        w: Width of the board
+        ha: First heuristic function to evaluate
+        hb: Second heuristic function to evaluate
+        alga: Search algorithm paired with ``ha``
+        algb: Search algorithm paired with ``hb``
+        num_iters: Number of iterations to average
+        kwargsa: Keyword args for only ``alga``
+        kwargsb: Keyword args for only ``algb``
+        kwargs: Keyword args for both algorithms
+
     Returns:
-        A tuple containing (avg. A, avg. B).
+        A tuple containing (avg. generated A, avg. generated B).
     """
+    if not kwargsa:
+        kwargsa = {}
+    if not kwargsb:
+        kwargsb = {}
     total_a, total_b = 0, 0
     for _ in range(num_iters):
         board = new_board(h, w)
         shuffle_board(board)
-        result = search(board, alg=alga, heuristic=ha, **kwargs)
-        total_a += result.expanded
-        result = search(board, alg=algb, heuristic=hb, **kwargs)
-        total_b += result.expanded
+        result = search(board, alg=alga, heuristic=ha, **kwargsa, **kwargs)
+        total_a += result.generated
+        result = search(board, alg=algb, heuristic=hb, **kwargsb, **kwargs)
+        total_b += result.generated
     return total_a / num_iters, total_b / num_iters
