@@ -19,18 +19,24 @@ Each function accepts a puzzle board as input, and returns a numeric value
 that indicates the estimated distance from the goal.
 """
 
+from typing import Callable, TypeAlias
+
 import math
 
 import slidingpuzzle
+from slidingpuzzle.board import Board, freeze_board
 
 
-def euclidean_distance(board: tuple[list[int], ...]) -> float:
+Heuristic: TypeAlias = Callable[[Board], int | float]
+
+
+def euclidean_distance(board: Board) -> float:
     r"""
     The distance between each tile and its destination, as measured in Euclidean space
 
     .. math::
-        \sum_{i}^{n} \sqrt{(\\text{tile}_{i, x} - \\text{goal}_{i, x})^2 + \\
-            (\\text{tile}_{i, y} - \\text{goal}_{i, y})^2}
+        \sum_{i}^{n} \sqrt{(\text{tile}_{i, x} - \text{goal}_{i, x})^2 +
+            (\text{tile}_{i, y} - \text{goal}_{i, y})^2}
 
     Args:
         board: The board
@@ -39,7 +45,7 @@ def euclidean_distance(board: tuple[list[int], ...]) -> float:
         The sum of all tile distances from their goal positions
     """
     w = len(board[0])
-    dist = 0
+    dist = 0.0
     for y, row in enumerate(board):
         for x, tile in enumerate(row):
             if slidingpuzzle.EMPTY_TILE == tile:
@@ -50,12 +56,12 @@ def euclidean_distance(board: tuple[list[int], ...]) -> float:
     return dist
 
 
-def hamming_distance(board: tuple[list[int], ...]) -> int:
+def hamming_distance(board: Board) -> int:
     r"""
     The count of misplaced tiles.
 
     .. math::
-        \sum_{i}^{n} \\text{tile}_i \oplus \\text{goal}_i
+        \sum_{i}^{n} \text{tile}_i \oplus \text{goal}_i
 
     Args:
         board: The board
@@ -76,14 +82,14 @@ def hamming_distance(board: tuple[list[int], ...]) -> int:
     return dist
 
 
-def manhattan_distance(board: tuple[list[int], ...]) -> int:
+def manhattan_distance(board: Board) -> int:
     r"""
     The minimum number of moves needed to restore the board to the goal state, if tiles
     could be moved through each other.
 
     .. math::
-        \sum_{i}^{n} |\\text{tile}_{i, x} - \\text{goal}_{i, x}| + \\
-            |\\text{tile}_{i, y} - \\text{goal}_{i, y}|
+        \sum_{i}^{n} |\text{tile}_{i, x} - \text{goal}_{i, x}| +
+            |\text{tile}_{i, y} - \text{goal}_{i, y}|
 
     Args:
         board: The board
@@ -103,12 +109,12 @@ def manhattan_distance(board: tuple[list[int], ...]) -> int:
     return dist
 
 
-def random_distance(board: tuple[list[int], ...]) -> int:
+def random_distance(board: Board) -> int:
     r"""
     A random distance computed as a hash of the board state. Useful as a baseline.
 
     .. math::
-        hash(\\text{board})
+        hash(\text{board})
 
     Args:
         board: The board
@@ -116,5 +122,4 @@ def random_distance(board: tuple[list[int], ...]) -> int:
     Returns:
         A random distance that is consistent for a given board state
     """
-    frozen_board = tuple(tuple(row) for row in board)
-    return hash(frozen_board)
+    return hash(freeze_board(board))
